/* 
  This program designed to train to destignguish two different
  digits dig1 and dig2. Also is can be used to destinguish one digit dig1 from all others
  with mentioning preprocess directive FROM_OTHERS */

#undef PLOT

#include <adolc.h>
#include <math.h>
#include <stdio.h>
#define athresh (0.2)

#define FOR(i,n) for (int (i)=0; (i)<(n); (i)+=1)


#define MAXNCYCLES 10000		/* max number of cycles through training data */
#define Ninp (2)
#define Nhid (2)				/* # of units in hidden layer */
#define Nout (1)				/* # of units in the output */

#define NA 30					// # of patterns in set A
#define NB 30					// # of patterns in set B
#define NC 30					// # of patterns in set B


extern "C" {
  // #include "meg/smd.h"
  // #include "meg/stochastic-gradient.h"
#include "meg/random.h"
}

#define  Nlev1 (Ninp*Nhid)		// number of weights input-hidden layers
#define  Nlev2 (Nhid*Nout)		// --//-- hidden-output layers

double wi[Nlev1+Nlev2], dwi[Nlev1+Nlev2];	/* weights inp-h*/
//  wh[Nlev2], dwh[Nlev2];

adoublev awi(Nlev1+Nlev2);	/* weights inp-h*/
//  awh (Nlev2);

double   x[Ninp]; 	        /* inputs */
adoublev  ax(Ninp);

adoublev ahid(Nhid);
double hid [Nhid];
adoublev aout(Nout);
double out [Nout];

adouble asum, err, mE;


adoublem AB (NA+NB,Ninp);
adoublem C  (NC,Ninp);

adoublem targets(NA+NB,Nout); 
adoublem targetC(NC,Nout); 

adouble sigmoid(adouble & x) {
  return (1 / (1 + exp(-x)));
}


void forward_pass(adoublev &x ){//adoublev &ahid) {
  int ih,io,in;

  asum=0;
  
  FOR (ih, Nhid) {
	FOR (in, Ninp)
	  asum += ax[in] * awi[in*Nhid+ih];
	ahid[ih] = sigmoid(asum);
  }
  
  FOR (io, Nout) {
	asum=0;
	FOR (ih, Nhid)
	  asum += ahid[ih] * awi[Nlev1+ih*Nout+io];
	aout[io] = sigmoid(asum);
	//	aout[io] = out[io];
  }
  
}

/* Error on current input with specified target */
adouble ErrorOut (adoublev & target) {
  adouble eres=0;
  FOR (i, Nout)
	eres += pow(aout[i] - target[i],2);
  return eres;
}

/* Error on one pass */
adouble ErrorPass(adoublev & xin, adoublev & target) {
  //  ConvertInput(xim, ax);
  forward_pass(xin);
  return (ErrorOut(target));
}

adouble Errors(adoublem & xin, adoublem & target,int starti, int endi)
{
  adouble Err=0;
  for (int i=starti; i<=endi; i++)
	Err += ErrorPass(xin[i], target[i]);

  return (Err/(endi-starti+1));
}

double oneStep(adoublev & in, adoublev & target, short  tag) {
  double E;
  volatile double eta = 0.001;
  trace_on(tag);
  awi <<= wi;
  //  awh <<= wh;
   
  ErrorPass(in, target) >>= E;
  trace_off();
#ifdef DEBUG
  {
	int counts[11];

	tapestats(tag, counts);
	cout << "# of independents =" << counts[0] << endl;
	cout << "# of dependents =" << counts[1] << endl;
	cout << "# of live active variables =" << counts[2] << endl;
	cout << "size of value stack =" << counts[3] << endl;
	cout << "buffer size =" << counts[4] << endl;
	cout << "# of operations recorded =" << counts[5] << endl;
  }
#endif
  gradient(tag, Nlev1+Nlev2, wi, dwi);
  //  gradient(tag, Nlev2, wh, dwh);
  
  /*update weights*/
  for(int i=0; i<Nlev1+Nlev2; i++)
	wi[i] -= eta * dwi[i];
  return E;
}

int main () {
  //  tdbase training, testing;
  unsigned int Nt,Mt,		/* number of training and testing images */
    i,j,k, hm=0;
  
  short  tag = 0;
  float EA=0, EB=0, prevEA=0, prevEB=0, EC=0;
  
#ifdef PLOT
  char *  comml1="../plotonline --title \"Training Errors\" -x Step  --no-point "; 
  char *  comml2="../plotonline --title \"Generalization Errors\" -x Step  --no-point "; 
  FILE *plotout1, *plotout2;            /* For plotonline pipe   */
#endif
  x[0]=1;
  
  cout << "Creating testing sets....." << endl;

  randomize_init();
  FOR(i,NA) {
	FOR(k,Ninp)
	  AB[i][k]=random_gauss(0, 1);
	FOR(k,Nout)
	  targets[i][k]=random_gauss(0.4, 0.2);
  }

  FOR(i,NB) {
	FOR(k,Ninp)
	  AB[i+NA][k]=random_gauss(1.5, 1);
	FOR(k,Nout)
	  targets[i+NA][k]=random_gauss(0.3, 0.2);
  }

  
  FOR(i,NC) {
	FOR(k,Ninp)
	  C[i][k]=random_gauss(0.5, 2);
	FOR(k,Nout)
	  targetC[i][k]=random_gauss(0.5, 0.2);
  }
  printf("Training...........\n");
  
  /* open a pipe for plot */
#ifdef PLOT
  plotout1 = popen(comml1,"w");
  plotout2 = popen(comml2,"w");
#endif
  
  randomize_gauss((double *)wi, Nlev1+Nlev2, 0, 0.7);

  fprintf(stderr,"$ DATA=COLUMN\n");
  fprintf(stderr," X A B\n");

  for (int step=0; step<4; step++) {
	prevEA=0; prevEB=0;
	for (j=0; j<MAXNCYCLES;j++) {
	  double trainE=0;
	  prevEA=EA;	prevEB=EB;
	  EA=0; EB=0;
	  for(int i=0; i < ((step==0)? NA+NB: NA ); i++) {
		if (!(i%100)) printf("Iter # %d  set pattern # %d\r", j, i);
		fflush(stdout);
		(i<NA? EA :EB) += oneStep(AB[i], targets[i], tag);
	  }
	  
	  EA /=  NA;
	  if (step==0) {
		EB /=  NB;
	  }
	  else {
		//EA = value(Errors(AB,targets,0,NA-1));//trainE/NA;
		EB = value(Errors(AB,targets,NA,NA+NB-1));
	  }
	  
	  EC = value(Errors(C,targetC,0,NC-1));

	  printf("Errors A=%3.5f  B=%3.5f C=%3.5f                                 \n",
			 EA, EB, 	EC);
	  fprintf(stderr,"%d %3.5f %3.5f %3.5f \n",hm++, EA, EB, EC);
	  
	  if ((prevEA) && (fabs(prevEA-EA)<0.0001*EA) &&
		  (prevEB) && (fabs(prevEB-EB)<0.0001*EB)) break;
	}

	/* Let's add some noice to weights */
	randomize_gauss((double *)dwi, Nlev1+Nlev2, 0, .5);
	for (int i=0; i<Nlev1+Nlev2; i++)
	  wi[i] += dwi[i];
  }
  

  return 0;
}
